/*
 * Copyright (c) 1999 Greg Haerr <greg@censoft.com>
 * Copyright (c) 1991 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Graphics server event routines for windows.
 */
#include "windows.h"
#include "wintern.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static int abs(int n);

/*
 * Update mouse status and issue events on it if necessary.
 * This function doesn't block, but is normally only called when
 * there is known to be some data waiting to be read from the mouse.
 */
BOOL
GsCheckMouseEvent(void)
{
	COORD		rootx;		/* latest mouse x position */
	COORD		rooty;		/* latest mouse y position */
	BUTTON		newbuttons;	/* latest buttons */
	int		mousestatus;	/* latest mouse status */

	/* Read the latest mouse status: */
	mousestatus = GdReadMouse(&rootx, &rooty, &newbuttons);
	if(mousestatus < 0) {
		/*GsError(GR_ERROR_MOUSE_ERROR, 0);*/
		return FALSE;
	} else if(mousestatus) {	/* Deliver events as appropriate: */	
		GsHandleMouseStatus(rootx, rooty, newbuttons);
		return TRUE;
	}
	return FALSE;
}

/*
 * Update keyboard status and issue events on it if necessary.
 * This function doesn't block, but is normally only called when
 * there is known to be some data waiting to be read from the keyboard.
 */
BOOL
GsCheckKeyboardEvent(void)
{
	unsigned char	ch;		/* latest character */
	MODIFIER	modifiers;	/* latest modifiers */
	int		keystatus;	/* latest keyboard status */

	/* Read the latest keyboard status: */
	keystatus = GdReadKeyboard(&ch, &modifiers);
	if(keystatus < 0) {
		if(keystatus == -2)	/* special case for ESC pressed*/
			GsTerminate();
		/*GsError(GR_ERROR_KEYBOARD_ERROR, 0);*/
		return FALSE;
	} else if(keystatus) {		/* Deliver events as appropriate: */	
		GsDeliverKeyboardEvent(ch, modifiers);
		return TRUE;
	}
	return FALSE;
}

/*
 * Handle all mouse events.  These are mouse enter, mouse exit, mouse
 * motion, mouse position, button down, and button up.  This also moves
 * the cursor to the new mouse position and changes it shape if needed.
 */
void GsHandleMouseStatus(COORD newx, COORD newy, BUTTON newbuttons)
{
	BUTTON		changebuttons;	/* buttons that have changed */
	unsigned char	ch;		/* latest character */
	MODIFIER	modifiers;	/* latest modifiers */
	static BUTTON curbuttons;

	GdReadKeyboard(&ch, &modifiers); /* Read the modifiers */

	/*
	 * First, if the mouse has moved, then position the cursor to the
	 * new location, which will send mouse enter, mouse exit, focus in,
	 * and focus out events if needed.  Check here for mouse motion and
	 * mouse position events.
	 */
	if (newx != cursorx || newy != cursory) {
		GsMoveCursor(newx, newy);
		GsDeliverMouseEvent(newbuttons, 0, modifiers);
	}

	/*
	 * Next, generate a button up event if any buttons have been released.
	 */
	changebuttons = (curbuttons & ~newbuttons);
	if (changebuttons)
		GsDeliverMouseEvent(newbuttons, changebuttons, modifiers);

	/*
	 * Finally, generate a button down event if any buttons have been
	 * pressed.
	 */
	changebuttons = (~curbuttons & newbuttons);
	if (changebuttons)
		GsDeliverMouseEvent(newbuttons, changebuttons, modifiers);

	curbuttons = newbuttons;
}

/*
 * Translate and deliver hardware mouse message to proper window.
 */
void
WndTranslateMouseMessage(HWND hwnd,UINT msg,int hittest)
{
	POINT		pt;
	DWORD		tick;
	static UINT	lastmsg = 0;
	static HWND	lasthwnd;
	static DWORD	lasttick;
	static int	lastx, lasty;

	/* determine double click eligibility*/
	if(msg == WM_LBUTTONDOWN || msg == WM_RBUTTONDOWN) {
		tick = GetTickCount();
		if((hwnd->pClass->style & CS_DBLCLKS) &&
		    msg == lastmsg && hwnd == lasthwnd &&
		    tick - lasttick < DBLCLICKSPEED &&
		    abs(cursorx-lastx) < SYSMETRICS_CXDOUBLECLK &&
		    abs(cursory-lasty) < SYSMETRICS_CYDOUBLECLK)
			msg += (WM_LBUTTONDBLCLK - WM_LBUTTONDOWN);
		lastmsg = msg;
		lasthwnd = hwnd;
		lasttick = tick;
		lastx = cursorx;
		lasty = cursory;
	}

	/*
	 * We always send nc mouse message
	 * unlike Windows, for HTCLIENT default processing
	 */
	PostMessage(hwnd, msg + (WM_NCMOUSEMOVE-WM_MOUSEMOVE), hittest,
		MAKELONG(cursorx, cursory));

	/* then possibly send user mouse message*/
	if(hittest == HTCLIENT) {
		pt.x = cursorx;
		pt.y = cursory;
		ScreenToClient(hwnd, &pt);
		PostMessage(hwnd, msg, 0, MAKELONG(pt.x, pt.y));
	}
}

/*
 * Deliver a mouse button or motion event.
 */
void 
GsDeliverMouseEvent(BUTTON buttons, BUTTON changebuttons, MODIFIER modifiers)
{
	HWND	hwnd;
	int	hittest;
	UINT	msg;

	hwnd = GetCapture();
	if(!hwnd)
		hwnd = mousewp;
	hittest = SendMessage(hwnd, WM_NCHITTEST, 0, MAKELONG(cursorx,cursory));

	if(!changebuttons)
		WndTranslateMouseMessage(hwnd, WM_MOUSEMOVE, hittest);

	if(changebuttons & LBUTTON) {
		msg = (buttons&LBUTTON)? WM_LBUTTONDOWN: WM_LBUTTONUP;
		WndTranslateMouseMessage(hwnd, msg, hittest);
	}

	if(changebuttons & RBUTTON) {
		msg = (buttons&RBUTTON)? WM_RBUTTONDOWN: WM_RBUTTONUP;
		WndTranslateMouseMessage(hwnd, msg, hittest);
	}
}

/*
 * Deliver a keyboard event.
 */
void GsDeliverKeyboardEvent(UCHAR ch, MODIFIER modifiers)
{
	SendMessage(focuswp, WM_CHAR, (WPARAM)ch, 0L);
}

/*
 * Deliver a window expose event.
 * Most of the work is in calculating the update region
 * for better redraw look and feel, and then queuing a
 * WM_PAINT message to the window.
 */
void
GsDeliverExposureEvent(HWND wp, COORD x, COORD y, COORD width, COORD height)
{
	if (wp->unmapcount)
		return;

	WndUnionUpdateRegion(wp, x, y, width, height);
	PostMessage(wp, WM_PAINT, 0, 0L);
}

/*
 * Union the passed rectangle with the update region for the given window.
 */
void
WndUnionUpdateRegion(HWND wp, COORD x, COORD y, COORD width, COORD height)
{
	if (wp->unmapcount)
		return;

#if UPDATEREGIONS
#if NUPDATEREGIONS > 1
	if(wp->nRcUpdate >= NUPDATEREGIONS) {
		printf("\007too many rcupdate\n");
		return;
	}
#else
	/* x, y are delivered relative to window*/
	if(wp->nRcUpdate == 1) {
		RECT	rc;

		/* Union expose area with existing update rectangle*/
		rc = wp->RcUpdate[0];
		rc.left = min(x, rc.left);
		rc.top = min(y, rc.top);
		rc.right = max(x+width, rc.right);
		rc.bottom = max(y+height, rc.bottom);
		wp->RcUpdate[0] = rc;
	} else
#endif
		SetRect(&wp->RcUpdate[wp->nRcUpdate++], x, y, x+width,y+height);
#endif
}

/*
 * Move the cursor to the specified absolute screen coordinates.
 * The coordinates are that of the defined hot spot of the cursor.
 * The cursor's appearance is changed to that defined for the window
 * in which the cursor is moved to.  In addition, the window the
 * cursor is in is recalculated.
 */
void GsMoveCursor(COORD x, COORD y)
{
	/*
	 * Move the cursor only if necessary, offsetting it to
	 * place the hot spot at the specified coordinates.
	 */
	if (x != cursorx || y != cursory) {
		if(curcursor)
			GdMoveCursor(x - curcursor->cursor.hotx,
				y - curcursor->cursor.hoty);
		cursorx = x;
		cursory = y;
	}

	/*
	 * Now check to see which window the mouse is in and whether or
	 * not the cursor shape should be changed.
	 */
	GsCheckMouseWindow();
	GsCheckCursor();
}

/*
 * Check to see if the cursor shape is the correct shape for its current
 * location.  If not, its shape is changed.
 */
void GsCheckCursor(void)
{
	HWND 		wp;		/* window cursor is in */
	HCURSOR		cp;		/* cursor definition */

	/*
	 * Get the cursor at its current position, and if it is not the
	 * currently defined one, then set the new cursor.  However,
	 * if the window is currently captured, then leave it alone.
	 */
	wp = capturewp;
	if (wp == NULL)
		wp = mousewp;

	cp = wp->cursor;
	if (cp == curcursor)
		return;

	/*
	 * It needs redefining, so do it.
	 */
	curcursor = cp;
	GdMoveCursor(cursorx - cp->cursor.hotx, cursory - cp->cursor.hoty);
	GdSetCursor(&cp->cursor);
}

/*
 * Find the window which is currently visible for the specified coordinates.
 * This just walks down the window tree looking for the deepest mapped
 * window which contains the specified point.  If the coordinates are
 * off the screen, the root window is returned.
 */
HWND
WndFindVisibleWindow(COORD x, COORD y)
{
	HWND	wp;		/* current window */
	HWND	retwp;		/* returned window */

	wp = rootwp;
	retwp = wp;
	while (wp) {
		if (!wp->unmapcount &&
		    wp->winrect.left <= x && wp->winrect.top <= y &&
		    wp->winrect.right > x && wp->winrect.bottom > y) {
			retwp = wp;
			wp = wp->children;
			continue;
		}
		wp = wp->siblings;
	}
	return retwp;
}

/*
 * Check to see if the window the mouse is currently in has changed.
 */
void GsCheckMouseWindow(void)
{
	HWND	wp;

	/* Don't change if window drag or capture in progress*/
	wp = dragwp;
	if(!wp)
		wp = capturewp;
	if(!wp)
		wp = WndFindVisibleWindow(cursorx, cursory);
	mousewp = wp;
}

/*
 *  Copy dstsiz bytes, including nul, from src to dst.
 *  Return # bytes, excluding nul, copied.
 */
int
strzcpy(char *dst,const char *src,int dstsiz)
{
	int	cc = dstsiz;

	/* return 0 on NULL src*/
	if(!src)
		cc = dstsiz = 1;

	while(--dstsiz > 0) {
		if((*dst++ = *src++) == '\0')
			return cc - dstsiz - 1;
	}
	*dst = 0;
	return cc - dstsiz - 1;
}

static int
abs(int n)
{
	return n >= 0? n: -n;
}
